import datetime
from typing import ClassVar

# keep 'List' instead of 'list' for audiences to not break py310 with the memory backend
from typing import List  # noqa: UP035
from urllib.parse import urlparse

from flask import current_app

from canaille.backends.models import Model
from canaille.core.models import User


class Client(Model):
    """OpenID Connect client definition.

    Based on the :rfc:`OAuth 2.0 Dynamic Client Registration protocols <7591>` and the
    `OpenID Connect RP-Initiated Logout <https://openid.net/specs/openid-connect-rpinitiated-1_0.html>`_ specifications.
    """

    identifier_attribute: ClassVar[str] = "client_id"

    description: str | None = None

    @property
    def trusted(self) -> bool:
        """Trusted clients don't require to display the consent page to users.

        A client is trusted if its client_uri domain matches one of the domains
        listed in TRUSTED_DOMAINS configuration. Supports:
        - Exact match: "example.com" matches "example.com"
        - Subdomain match: "example.com" also matches "sub.example.com"
        - Wildcard match: ".example.com" matches "example.com" and all its subdomains
        """
        if not self.client_uri:
            return False

        parsed = urlparse(self.client_uri)
        hostname = parsed.hostname
        if not hostname:
            return False

        trusted_domains = current_app.config["CANAILLE_OIDC"]["TRUSTED_DOMAINS"]
        for domain in trusted_domains:
            # Wildcard match: .example.com matches example.com and all subdomains
            if domain.startswith("."):
                domain_without_dot = domain[1:]
                if hostname == domain_without_dot or hostname.endswith(domain):
                    return True

            # Exact match only for non-wildcard domains
            elif hostname == domain:
                return True

        return False

    # keep 'List' instead of 'list' do not break py310 with the memory backend
    audience: List["Client"] = []  # noqa: UP006

    client_id: str | None
    """REQUIRED.

    OAuth 2.0 client identifier string.  It SHOULD NOT be currently
    valid for any other registered client, though an authorization
    server MAY issue the same client identifier to multiple instances of
    a registered client at its discretion.
    """

    client_secret: str | None = None
    """OPTIONAL.

    OAuth 2.0 client secret string.  If issued, this MUST be unique for
    each "client_id" and SHOULD be unique for multiple instances of a
    client using the same "client_id".  This value is used by
    confidential clients to authenticate to the token endpoint, as
    described in OAuth 2.0 [RFC6749], Section 2.3.1.
    """

    client_id_issued_at: datetime.datetime | None = None
    """OPTIONAL.

    Time at which the client identifier was issued.  The time is
    represented as the number of seconds from 1970-01-01T00:00:00Z as
    measured in UTC until the date/time of issuance.
    """

    client_secret_expires_at: datetime.datetime | None = None
    """REQUIRED if "client_secret" is issued.

    Time at which the client secret will expire or 0 if it will not
    expire.  The time is represented as the number of seconds from
    1970-01-01T00:00:00Z as measured in UTC until the date/time of
    expiration.
    """

    redirect_uris: list[str] = []
    """Array of redirection URI strings for use in redirect-based flows such as
    the authorization code and implicit flows.

    As required by Section 2 of OAuth 2.0 [RFC6749], clients using flows
    with redirection MUST register their redirection URI values.
    Authorization servers that support dynamic registration for
    redirect-based flows MUST implement support for this metadata value.
    """

    token_endpoint_auth_method: str | None = None
    """String indicator of the requested authentication method for the token
    endpoint.  Values defined by this specification are:

    *  "none": The client is a public client as defined in OAuth 2.0,
       Section 2.1, and does not have a client secret.

    *  "client_secret_post": The client uses the HTTP POST parameters
       as defined in OAuth 2.0, Section 2.3.1.

    *  "client_secret_basic": The client uses HTTP Basic as defined in
       OAuth 2.0, Section 2.3.1.

    Additional values can be defined via the IANA "OAuth Token
    Endpoint Authentication Methods" registry established in
    Section 4.2.  Absolute URIs can also be used as values for this
    parameter without being registered.  If unspecified or omitted,
    the default is "client_secret_basic", denoting the HTTP Basic
    authentication scheme as specified in Section 2.3.1 of OAuth 2.0.
    """

    token_endpoint_auth_signing_alg: str | None = None
    """JWS [JWS] alg algorithm [JWA] that MUST be used for signing the JWT [JWT]
    used to authenticate the Client at the Token Endpoint for the private_key_jwt
    and client_secret_jwt authentication methods. All Token Requests using these
    authentication methods from this Client MUST be rejected, if the JWT is not
    signed with this algorithm. Servers SHOULD support RS256. The value none MUST
    NOT be used. The default, if omitted, is that any algorithm supported by the
    OP and the RP MAY be used."""

    grant_types: list[str] = ["authorization_code", "refresh_token"]
    """Array of OAuth 2.0 grant type strings that the client can use at the
    token endpoint.  These grant types are defined as follows:

    *  "authorization_code": The authorization code grant type defined
       in OAuth 2.0, Section 4.1.

    *  "implicit": The implicit grant type defined in OAuth 2.0,
       Section 4.2.

    *  "password": The resource owner password credentials grant type
       defined in OAuth 2.0, Section 4.3.

    *  "client_credentials": The client credentials grant type defined
       in OAuth 2.0, Section 4.4.

    *  "refresh_token": The refresh token grant type defined in OAuth
       2.0, Section 6.

    *  "urn:ietf:params:oauth:grant-type:jwt-bearer": The JWT Bearer
       Token Grant Type defined in OAuth JWT Bearer Token Profiles
       [RFC7523].

    *  "urn:ietf:params:oauth:grant-type:saml2-bearer": The SAML 2.0
       Bearer Assertion Grant defined in OAuth SAML 2 Bearer Token
       Profiles [RFC7522].

    If the token endpoint is used in the grant type, the value of this
    parameter MUST be the same as the value of the "grant_type"
    parameter passed to the token endpoint defined in the grant type
    definition.  Authorization servers MAY allow for other values as
    defined in the grant type extension process described in OAuth
    2.0, Section 4.5.  If omitted, the default behavior is that the
    client will use only the "authorization_code" Grant Type.
    """

    response_types: list[str] = []
    """
    Array of the OAuth 2.0 response type strings that the client can
    use at the authorization endpoint.  These response types are
    defined as follows:

    *  "code": The authorization code response type defined in OAuth
       2.0, Section 4.1.

    *  "token": The implicit response type defined in OAuth 2.0,
       Section 4.2.

    If the authorization endpoint is used by the grant type, the value
    of this parameter MUST be the same as the value of the
    "response_type" parameter passed to the authorization endpoint
    defined in the grant type definition.  Authorization servers MAY
    allow for other values as defined in the grant type extension
    process is described in OAuth 2.0, Section 4.5.  If omitted, the
    default is that the client will use only the "code" response type.
    """

    client_name: str | None = None
    """Human-readable string name of the client to be presented to the end-user
    during authorization.

    If omitted, the authorization server MAY display the raw "client_id"
    value to the end-user instead.  It is RECOMMENDED that clients
    always send this field. The value of this field MAY be
    internationalized, as described in Section 2.2.
    """

    client_uri: str | None = None
    """URL string of a web page providing information about the client.

    If present, the server SHOULD display this URL to the end-user in a
    clickable fashion.  It is RECOMMENDED that clients always send this
    field.  The value of this field MUST point to a valid web page.  The
    value of this field MAY be internationalized, as described in
    Section 2.2.
    """

    logo_uri: str | None = None
    """URL string that references a logo for the client.

    If present, the server SHOULD display this image to the end-user
    during approval. The value of this field MUST point to a valid image
    file.  The value of this field MAY be internationalized, as
    described in Section 2.2.
    """

    scope: list[str] = []
    """String containing a space-separated list of scope values (as described
    in Section 3.3 of OAuth 2.0 [RFC6749]) that the client can use when
    requesting access tokens.

    The semantics of values in this list are service specific.  If
    omitted, an authorization server MAY register a client with a
    default set of scopes.
    """

    contacts: list[str] = []
    """Array of strings representing ways to contact people responsible for
    this client, typically email addresses.

    The authorization server MAY make these contact addresses available
    to end-users for support requests for the client.  See Section 6 for
    information on Privacy Considerations.
    """

    tos_uri: str | None = None
    """URL string that points to a human-readable terms of service document for
    the client that describes a contractual relationship between the end-user
    and the client that the end-user accepts when authorizing the client.

    The authorization server SHOULD display this URL to the end-user if
    it is provided.  The value of this field MUST point to a valid web
    page.  The value of this field MAY be internationalized, as
    described in Section 2.2.
    """

    policy_uri: str | None = None
    """URL string that points to a human-readable privacy policy document that
    describes how the deployment organization collects, uses, retains, and
    discloses personal data.

    The authorization server SHOULD display this URL to the end-user if
    it is provided.  The value of this field MUST point to a valid web
    page.  The value of this field MAY be internationalized, as
    described in Section 2.2.
    """

    jwks_uri: str | None = None
    """URL string referencing the client's JSON Web Key (JWK) Set [RFC7517]
    document, which contains the client's public keys.

    The value of this field MUST point to a valid JWK Set document.
    These keys can be used by higher-level protocols that use signing or
    encryption.  For instance, these keys might be used by some
    applications for validating signed requests made to the token
    endpoint when using JWTs for client authentication [RFC7523].  Use
    of this parameter is preferred over the "jwks" parameter, as it
    allows for easier key rotation.  The "jwks_uri" and "jwks"
    parameters MUST NOT both be present in the same request or response.
    """

    jwks: str | None = None
    """Client's JSON Web Key Set [RFC7517] document value, which contains the
    client's public keys.

    The value of this field MUST be a JSON object containing a valid JWK
    Set.  These keys can be used by higher-level protocols that use
    signing or encryption.  This parameter is intended to be used by
    clients that cannot use the "jwks_uri" parameter, such as native
    clients that cannot host public URLs.  The "jwks_uri" and "jwks"
    parameters MUST NOT both be present in the same request or response.
    """

    sector_identifier_uri: str | None = None
    """URL using the https scheme to be used in calculating Pseudonymous Identifiers by the OP.
    The URL references a file with a single JSON array of redirect_uri values.
    Please see Section 5. Providers that use pairwise sub (subject) values SHOULD
    utilize the sector_identifier_uri value provided in the Subject Identifier
    calculation for pairwise identifiers."""

    subject_type: str | None = None
    """subject_type requested for responses to this Client.
    The subject_types_supported discovery parameter contains a
    list of the supported subject_type values for the OP.
    Valid types include pairwise and public."""

    software_id: str | None = None
    """A unique identifier string (e.g., a Universally Unique Identifier
    (UUID)) assigned by the client developer or software publisher used by
    registration endpoints to identify the client software to be dynamically
    registered.

    Unlike "client_id", which is issued by the authorization server and
    SHOULD vary between instances, the "software_id" SHOULD remain the
    same for all instances of the client software.  The "software_id"
    SHOULD remain the same across multiple updates or versions of the
    same piece of software.  The value of this field is not intended to
    be human readable and is usually opaque to the client and
    authorization server.
    """

    software_version: str | None = None
    """A version identifier string for the client software identified by
    "software_id".

    The value of the "software_version" SHOULD change on any update to
    the client software identified by the same "software_id".  The value
    of this field is intended to be compared using string equality
    matching and no other comparison semantics are defined by this
    specification.  The value of this field is outside the scope of this
    specification, but it is not intended to be human readable and is
    usually opaque to the client and authorization server.  The
    definition of what constitutes an update to client software that
    would trigger a change to this value is specific to the software
    itself and is outside the scope of this specification.
    """

    post_logout_redirect_uris: list[str] = []
    """Array of URLs supplied by the RP to which it MAY request that the
    End-User's User Agent be redirected using the
    post_logout_redirect_uri parameter after a logout has been
    performed.
    These URLs SHOULD use the https scheme and MAY contain
    port, path, and query parameter components; however, they MAY use
    the http scheme, provided that the Client Type is confidential, as
    defined in Section 2.1 of OAuth 2.0 [RFC6749], and provided the OP
    allows the use of http RP URIs.
    """

    application_type: str = "web"
    """Kind of the application.
    The default, if omitted, is web. The defined values are native or web.
    Web Clients using the OAuth Implicit Grant Type MUST only register URLs
    using the https scheme as redirect_uris; they MUST NOT use localhost as
    the hostname. Native Clients MUST only register redirect_uris using custom
    URI schemes or loopback URLs using the http scheme; loopback URLs use
    localhost or the IP loopback literals 127.0.0.1 or [::1] as the hostname.
    Authorization Servers MAY place additional constraints on Native Clients.
    Authorization Servers MAY reject Redirection URI values using the http scheme,
    other than the loopback case for Native Clients. The Authorization Server
    MUST verify that all the registered redirect_uris conform to these constraints.
    This prevents sharing a Client ID across different types of Clients."""

    id_token_signed_response_alg: str = "RS256"
    """JWS alg algorithm [JWA] REQUIRED for signing the ID Token issued to this Client.
    The value none MUST NOT be used as the ID Token alg value unless the Client uses
    only Response Types that return no ID Token from the Authorization Endpoint
    (such as when only using the Authorization Code Flow). The default, if omitted, is RS256.
    The public key for validating the signature is provided by retrieving the JWK Set referenced
    by the jwks_uri element from OpenID Connect Discovery 1.0 [OpenID.Discovery]."""

    id_token_encrypted_response_alg: str | None = None
    """JWE alg algorithm [JWA] REQUIRED for encrypting the ID Token issued to this Client.
    If this is requested, the response will be signed then encrypted,
    with the result being a Nested JWT, as defined in [JWT]. The default, if omitted,
    is that no encryption is performed."""

    id_token_encrypted_response_enc: str | None = None
    """JWE enc algorithm [JWA] REQUIRED for encrypting the ID Token issued to this Client.
    If id_token_encrypted_response_alg is specified, the default id_token_encrypted_response_enc
    value is A128CBC-HS256. When id_token_encrypted_response_enc is included,
    id_token_encrypted_response_alg MUST also be provided."""

    userinfo_signed_response_alg: str | None = None
    """JWS alg algorithm [JWA] REQUIRED for signing UserInfo Responses.
    If this is specified, the response will be JWT [JWT] serialized, and signed using JWS.
    The default, if omitted, is for the UserInfo Response to return the
    Claims as a UTF-8 [RFC3629] encoded JSON object using the application/json content-type."""

    userinfo_encrypted_response_alg: str | None = None
    """JWE [JWE] alg algorithm [JWA] REQUIRED for encrypting UserInfo Responses. If both
    signing and encryption are requested, the response will be signed then encrypted,
    with the result being a Nested JWT, as defined in [JWT]. The default, if omitted,
    is that no encryption is performed."""

    userinfo_encrypted_response_enc: str | None = None
    """JWE enc algorithm [JWA] REQUIRED for encrypting UserInfo Responses.
    If userinfo_encrypted_response_alg is specified, the default userinfo_encrypted_response_enc
    value is A128CBC-HS256. When userinfo_encrypted_response_enc is included,
    userinfo_encrypted_response_alg MUST also be provided."""

    default_max_age: int | None = None
    """Default Maximum Authentication Age.
    Specifies that the End-User MUST be actively authenticated if the End-User was authenticated
    longer ago than the specified number of seconds. The max_age request parameter overrides
    this default value. If omitted, no default Maximum Authentication Age is specified."""

    require_auth_time: bool = False
    """Boolean value specifying whether the auth_time Claim in the ID Token is REQUIRED.
    It is REQUIRED when the value is true. (If this is false, the auth_time Claim can still
    be dynamically requested as an individual Claim for the ID Token using the claims request
    parameter described in Section 5.5.1 of OpenID Connect Core 1.0 [OpenID.Core].)
    If omitted, the default value is false."""

    default_acr_values: list[str] = []
    """Default requested Authentication Context Class Reference values.
    Array of strings that specifies the default acr values that the OP is being requested
    to use for processing requests from this Client, with the values appearing in order of
    preference. The Authentication Context Class satisfied by the authentication performed
    is returned as the acr Claim Value in the issued ID Token. The acr Claim is requested
    as a Voluntary Claim by this parameter. The acr_values_supported discovery element
    contains a list of the supported acr values supported by the OP. Values specified in
    the acr_values request parameter or an individual acr Claim request override these
    default values."""

    initiate_login_uri: str | None = None
    """URI using the https scheme that a third party can use to initiate a login by the RP,
    as specified in Section 4 of OpenID Connect Core 1.0 [OpenID.Core]. The URI MUST accept
    requests via both GET and POST. The Client MUST understand the login_hint and iss parameters
    and SHOULD support the target_link_uri parameter."""

    request_object_signing_alg: str | None = None
    """JWS [JWS] alg algorithm [JWA] that MUST be used for signing Request Objects sent to the OP.
    All Request Objects from this Client MUST be rejected, if not signed with this algorithm.
    Request Objects are described in Section 6.1 of OpenID Connect Core 1.0 [OpenID.Core].
    This algorithm MUST be used both when the Request Object is passed by value (using the request parameter)
    and when it is passed by reference (using the request_uri parameter). Servers SHOULD support RS256.
    The value none MAY be used. The default, if omitted, is that any algorithm supported by the
    OP and the RP MAY be used."""

    request_object_encryption_alg: str | None = None
    """JWE [JWE] alg algorithm [JWA] the RP is declaring that it may use for encrypting Request Objects sent to the OP.
    This parameter SHOULD be included when symmetric encryption will be used, since this signals to the OP
    that a client_secret value needs to be returned from which the symmetric key will be derived,
    that might not otherwise be returned. The RP MAY still use other supported encryption algorithms
    or send unencrypted Request Objects, even when this parameter is present. If both signing and
    encryption are requested, the Request Object will be signed then encrypted, with the result
    being a Nested JWT, as defined in [JWT]. The default, if omitted, is that the RP is not declaring
    whether it might encrypt any Request Objects."""

    request_object_encryption_enc: str | None = None
    """JWE enc algorithm [JWA] the RP is declaring that it may use for encrypting Request Objects sent to the OP.
    If request_object_encryption_alg is specified, the default request_object_encryption_enc value is A128CBC-HS256.
    When request_object_encryption_enc is included, request_object_encryption_alg MUST also be provided."""

    request_uris: list[str] = []
    """Array of request_uri values that are pre-registered by the RP for use at the OP.
    These URLs MUST use the https scheme unless the target Request Object is signed in a way
    that is verifiable by the OP. Servers MAY cache the contents of the files referenced by these
    URIs and not retrieve them at the time they are used in a request.
    OPs can require that request_uri values used be pre-registered with the
    require_request_uri_registration discovery parameter.
    If the contents of the request file could ever change, these URI values SHOULD include
    the base64url-encoded SHA-256 hash value of the file contents referenced by the URI as
    the value of the URI fragment. If the fragment value used for a URI changes,
    that signals the server that its cached value for that URI with the old fragment
    value is no longer valid."""

    require_signed_request_object: bool = False
    """Indicates where authorization request needs to be protected as Request Object and provided
    through either request or request_uri parameter."""


class AuthorizationCode(Model):
    """OpenID Connect temporary authorization code definition."""

    identifier_attribute: ClassVar[str] = "authorization_code_id"

    authorization_code_id: str
    code: str
    client: "Client"
    subject: User
    redirect_uri: str | None
    response_type: str | None
    scope: list[str]
    nonce: str | None
    issue_date: datetime.datetime
    lifetime: int
    challenge: str | None
    challenge_method: str | None
    revokation_date: datetime.datetime

    auth_time: datetime.datetime
    """Time when the End-User authentication occurred."""

    acr: str | None
    """Authentication Context Class Reference.
    String specifying an Authentication Context Class Reference value that identifies
    the Authentication Context Class that the authentication performed satisfied."""

    amr: list[str]
    """Authentication Methods References.
    JSON array of strings that are identifiers for authentication methods used in
    the authentication."""


class Token(Model):
    """OpenID Connect token definition."""

    identifier_attribute: ClassVar[str] = "token_id"

    token_id: str
    access_token: str
    client: "Client"
    subject: User | None
    type: str
    refresh_token: str
    scope: list[str]
    issue_date: datetime.datetime
    lifetime: int
    revokation_date: datetime.datetime
    # keep 'List' instead of 'list' do not break py310 with the memory backend
    audience: List["Client"]  # noqa: UP006


class Consent(Model):
    """Long-term user consent to an application."""

    identifier_attribute: ClassVar[str] = "consent_id"

    consent_id: str
    subject: User
    client: "Client"
    scope: list[str]
    issue_date: datetime.datetime
    revokation_date: datetime.datetime

    def revoke(self):
        raise NotImplementedError()

    def restore(self):
        raise NotImplementedError()



import logging
from unittest import mock

import time_machine
from flask import current_app
from flask import url_for

from canaille.app import models
from canaille.core.endpoints.account import RegistrationPayload


def test_registration_without_email_validation(testclient, backend, foo_group):
    """Tests a nominal registration without email validation."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")
    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm a little pea"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"
    res = res.form.submit()
    assert ("success", "Your account has been created successfully.") in res.flashes

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


def test_registration_failure_with_different_passwords_and_too_short_password(
    testclient, backend, foo_group
):
    """Tests a nominal registration without email validation but with a wrong confirmation password and a too short password."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")
    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "123"
    res.form["password2"] = "123"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"
    res = res.form.submit()
    assert ("error", "User account creation failed.") in res.flashes
    res.mustcontain("Field must be at least 8 characters long.")

    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm not a little pea"
    res = res.form.submit()
    res.mustcontain("Password and confirmation do not match.")
    res.mustcontain('data-percent="100"')

    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm a little pea"
    res = res.form.submit()
    assert ("success", "Your account has been created successfully.") in res.flashes

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


def test_registration_with_email_validation(testclient, backend, smtpd, foo_group):
    """Tests a nominal registration with email validation."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True

    with time_machine.travel("2020-01-01 02:00:00+00:00", tick=False):
        res = testclient.get(url_for("core.account.join"))
        res.form["email"] = "foo@bar.test"
        res = res.form.submit()

    assert res.flashes == [
        (
            "success",
            "You will receive soon an email to continue the registration process.",
        )
    ]
    assert len(smtpd.messages) == 1

    payload = RegistrationPayload(
        creation_date_isoformat="2020-01-01T02:00:00+00:00",
        user_name="",
        user_name_editable=True,
        email="foo@bar.test",
        groups=[],
    )
    registration_url = url_for(
        "core.account.registration",
        data=payload.b64(),
        hash=payload.build_hash(),
        _external=True,
    )
    text_mail = smtpd.messages[0].get_payload()[0].get_payload(decode=True).decode()
    assert registration_url in text_mail

    assert not backend.query(models.User, user_name="newuser")
    with time_machine.travel("2020-01-01 02:01:00+00:00", tick=False):
        res = testclient.get(registration_url, status=200)
        res.form["user_name"] = "newuser"
        res.form["password1"] = "i'm a little pea"
        res.form["password2"] = "i'm a little pea"
        res.form["family_name"] = "newuser"
        res = res.form.submit()

    assert res.flashes == [
        ("success", "Your account has been created successfully."),
    ]

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


def test_registration_with_email_already_taken(
    testclient, backend, smtpd, user, foo_group
):
    """Be sure to not leak email existence if HIDE_INVALID_LOGINS is true."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = True
    res = testclient.get(url_for("core.account.join"))
    res.form["email"] = "john@doe.test"
    res = res.form.submit()
    assert res.flashes == [
        (
            "success",
            "You will receive soon an email to continue the registration process.",
        )
    ]

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = False
    res = testclient.get(url_for("core.account.join"))
    res.form["email"] = "john@doe.test"
    res = res.form.submit()
    assert res.flashes == []
    res.mustcontain("The email 'john@doe.test' is already used")


def test_registration_with_email_validation_needs_a_valid_link(
    testclient, backend, smtpd, foo_group
):
    """Tests a nominal registration without email validation."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.get(url_for("core.account.registration"), status=403)


def test_join_page_registration_disabled(testclient, backend, smtpd, foo_group):
    """The join page should not be available if registration is disabled."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = False
    testclient.get(url_for("core.account.join"), status=404)


def test_join_page_email_confirmation_disabled(testclient, backend, smtpd, foo_group):
    """The join page should directly redirect to the registration page if email confirmation is disabled."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False
    res = testclient.get(url_for("core.account.join"), status=302)
    assert res.location == url_for("core.account.registration")


def test_join_page_already_logged_in(testclient, backend, logged_user, foo_group):
    """The join page should not be accessible for logged users."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.get(url_for("core.account.join"), status=403)


@mock.patch("smtplib.SMTP")
def test_registration_mail_error(SMTP, testclient, backend, smtpd, foo_group):
    """Display an error message if the registration mail could not be sent."""
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    SMTP.side_effect = mock.Mock(side_effect=OSError("unit test mail error"))
    res = testclient.get(url_for("core.account.join"))
    res.form["email"] = "foo@bar.test"
    res = res.form.submit(expect_errors=True)

    assert res.flashes == [
        (
            "error",
            "An error happened while sending your registration mail. "
            "Please try again in a few minutes. "
            "If this still happens, please contact the administrators.",
        )
    ]
    assert len(smtpd.messages) == 0


@mock.patch("httpx.get")
def test_registration_with_compromised_password(api_get, testclient, backend):
    """Tests a nominal registration with compromised password."""
    current_app.config["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    # This content simulates a result from the hibp api containing the suffixes of the following password hashes: 'password', '987654321', 'correct horse battery staple', 'zxcvbn123', 'azertyuiop123'
    class Response:
        content = b"1E4C9B93F3F0682250B6CF8331B7EE68FD8:3\r\nCAA6D483CC3887DCE9D1B8EB91408F1EA7A:3\r\nAD6438836DBE526AA231ABDE2D0EEF74D42:3\r\n8289894DDB6317178960AB5AE98B81BBF97:1\r\n5FF0B6F9EAC40D5CA7B4DAA7B64F0E6F4AA:2\r\n"

    api_get.return_value = Response

    assert not backend.query(models.User, user_name="newuser")
    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "987654321"
    res.form["password2"] = "987654321"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"
    res = res.form.submit()
    res.mustcontain(
        "This password appears on public compromission databases and is not secure."
    )

    user = backend.get(models.User, user_name="newuser")
    assert user is None


@mock.patch("httpx.get")
def test_registration_with_compromised_password_request_api_failed_but_account_created(
    api_get, testclient, backend, caplog
):
    current_app.config["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    api_get.side_effect = mock.Mock(side_effect=Exception())
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")

    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"

    res = res.form.submit()

    assert (
        "canaille",
        logging.ERROR,
        "Password compromise investigation failed on HIBP API.",
    ) in caplog.record_tuples
    assert (
        "error",
        "Password compromise investigation failed. Please contact the administrators.",
    ) in res.flashes
    assert ("success", "Your account has been created successfully.") in res.flashes

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


@mock.patch("httpx.get")
def test_compromised_password_validator_with_failure_of_api_request_and_success_mail_to_admin_from_register_form(
    api_get, testclient, backend, caplog, smtpd
):
    current_app.config["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    api_get.side_effect = mock.Mock(side_effect=Exception())
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")

    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"

    res = res.form.submit()

    assert (
        "canaille",
        logging.ERROR,
        "Password compromise investigation failed on HIBP API.",
    ) in caplog.record_tuples
    assert (
        "error",
        "Password compromise investigation failed. Please contact the administrators.",
    ) in res.flashes
    assert (
        "info",
        "We have informed your administrator about the failure of the password compromise investigation.",
    ) in res.flashes
    assert ("success", "Your account has been created successfully.") in res.flashes
    assert len(smtpd.messages) == 1

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


@mock.patch("httpx.get")
def test_compromised_password_validator_with_failure_of_api_request_and_fail_to_send_mail_to_admin_from_register_form(
    api_get, testclient, backend, caplog
):
    current_app.config["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    api_get.side_effect = mock.Mock(side_effect=Exception())
    current_app.config["CANAILLE"]["SMTP"]["TLS"] = False
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")

    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"

    res = res.form.submit()

    assert (
        "canaille",
        logging.ERROR,
        "Password compromise investigation failed on HIBP API.",
    ) in caplog.record_tuples
    assert (
        "error",
        "Password compromise investigation failed. Please contact the administrators.",
    ) in res.flashes
    assert (
        "error",
        "An error occurred while communicating the incident to the administrators. "
        "Please update your password as soon as possible. "
        "If this still happens, please contact the administrators.",
    ) in res.flashes
    assert ("success", "Your account has been created successfully.") in res.flashes

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


@mock.patch("httpx.get")
def test_compromised_password_validator_with_failure_of_api_request_without_smtp_from_register_form(
    api_get, testclient, backend, caplog
):
    current_app.config["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    api_get.side_effect = mock.Mock(side_effect=Exception())
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")

    current_app.config["CANAILLE"]["SMTP"] = None

    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"

    res = res.form.submit()

    assert (
        "canaille",
        logging.ERROR,
        "Password compromise investigation failed on HIBP API.",
    ) in caplog.record_tuples
    assert (
        "error",
        "Password compromise investigation failed. Please contact the administrators.",
    ) not in res.flashes

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)


@mock.patch("httpx.get")
def test_compromised_password_validator_with_failure_of_api_request_without_admin_email_from_register_form(
    api_get, testclient, backend, caplog
):
    current_app.config["CANAILLE"]["ENABLE_PASSWORD_COMPROMISSION_CHECK"] = True
    api_get.side_effect = mock.Mock(side_effect=Exception())
    testclient.app.config["CANAILLE"]["ENABLE_REGISTRATION"] = True
    testclient.app.config["CANAILLE"]["EMAIL_CONFIRMATION"] = False

    assert not backend.query(models.User, user_name="newuser")

    current_app.config["CANAILLE"]["ADMIN_EMAIL"] = None

    res = testclient.get(url_for("core.account.registration"), status=200)
    res.form["user_name"] = "newuser"
    res.form["password1"] = "123456789"
    res.form["password2"] = "123456789"
    res.form["family_name"] = "newuser"
    res.form["emails-0"] = "newuser@example.test"

    res = res.form.submit()

    assert (
        "canaille",
        logging.ERROR,
        "Password compromise investigation failed on HIBP API.",
    ) in caplog.record_tuples
    assert (
        "error",
        "Password compromise investigation failed. Please contact the administrators.",
    ) not in res.flashes

    user = backend.get(models.User, user_name="newuser")
    assert user
    backend.delete(user)



import random

import faker
from faker.config import AVAILABLE_LOCALES

from canaille.app import models
from canaille.app.i18n import available_language_codes
from canaille.backends import Backend


def fake_users(nb=1):
    locales = list(set(available_language_codes()) & set(AVAILABLE_LOCALES))
    faker_obj = faker.Faker(locales)
    users = list()

    for _ in range(nb):
        try:
            locale = random.choice(locales)
            fake = faker_obj[locale]
            profile = fake.unique.profile()
            name = profile["name"]
            user = models.User(
                formatted_name=name,
                given_name=name.split(" ")[0],
                family_name=name.split(" ")[1],
                user_name=profile["username"],
                emails=[profile["mail"]],
                phone_numbers=[profile["ssn"]],
                profile_url=profile["website"][0],
                formatted_address=profile["address"],
                street=fake.street_name(),
                postal_code=fake.postcode(),
                locality=fake.city(),
                region=fake.state(),
                employee_number=str(fake.unique.random_number()),
                department=fake.word(),
                title=profile["job"],
                password=fake.password(),
                preferred_language=fake._locales[0],
            )
            Backend.instance.save(user)
            users.append(user)
        except Exception:  # pragma: no cover
            pass
    return users


def fake_groups(nb=1, nb_users_max=1):
    users = Backend.instance.query(models.User)
    groups = list()
    fake = faker.Faker(["en_US"])
    for _ in range(nb):
        try:
            group = models.Group(
                display_name=fake.unique.word(),
                description=fake.sentence(),
            )
            if nb_users_max:
                nb_users = random.randrange(1, nb_users_max + 1)
                group.members = list({random.choice(users) for _ in range(nb_users)})
            Backend.instance.save(group)
            groups.append(group)
        except Exception:  # pragma: no cover
            pass
    return groups



import re

from canaille.commands import cli


def test_check_command(cli_runner):
    res = cli_runner.invoke(cli, ["--version"])
    semver_pattern = r"(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
    assert re.match(rf"Canaille, version {semver_pattern}", res.stdout)
    assert res.exit_code == 0, res.stdout



import json

from canaille.app import models
from canaille.commands import cli


def test_restore_stdin(cli_runner, backend):
    """Test the full database dump command."""
    payload = {
        "client": [
            {
                "id": "45fb7f55-816a-41bf-a9f0-16bb1d60f448",
                "created": "2025-03-15T12:55:32+00:00",
                "last_modified": "2025-03-15T12:55:32+00:00",
                "audience": ["45fb7f55-816a-41bf-a9f0-16bb1d60f448"],
                "client_id": "dx8HWLJOM2GRQtkR92fjwHZ7",
                "client_secret": "1pwUfTGWJCsvPVJdoNyMZIJiQuUEO6KBX5HA49G8OTsAtdXY",
                "client_id_issued_at": "2025-03-15T12:55:32.745494+00:00",
                "redirect_uris": [
                    "https://client.trusted.test/redirect1",
                    "https://client.trusted.test/redirect2",
                ],
                "token_endpoint_auth_method": "client_secret_basic",
                "grant_types": [
                    "password",
                    "authorization_code",
                    "implicit",
                    "hybrid",
                    "refresh_token",
                    "client_credentials",
                    "urn:ietf:params:oauth:grant-type:jwt-bearer",
                ],
                "response_types": ["code", "token", "id_token"],
                "client_name": "Some other client",
                "client_uri": "https://client.trusted.test",
                "logo_uri": "https://client.trusted.test/logo.webp",
                "scope": ["openid", "profile", "groups"],
                "contacts": ["contact@trusted.test"],
                "tos_uri": "https://client.trusted.test/tos",
                "policy_uri": "https://client.trusted.test/policy",
                "jwks": '{"keys": [{"n": "02uEn7s1vpHLEwNRV-5mQn0W-N--KhPoFuJ28NoRuGGBVoFmCLxnd70qdqu_8tQRuu6R65566PGiDLavkbNqe0cFZB7VSk2vnhpfJGLzUo_FYQeSK-XafnfIiuhd1IdCG6-ck7YfKJMnfqpQW_hFnq-k1SXkZVhcEsYtZ-wtXfa3hiTPrT4oFzEf0gAAvluaTTKOXaGELo-8NDv4HycQPzWLSvpAqtAEyVo_pjfAUrdDTaNPtiQG1-ISrx3UKxji0Wmt8oaJyOABQ0y3dpvT5EyvWfkooOjlebK4dI-a8ZmSU6ev_x3-HWy-9W3wkWiBCscBgiR_vmd1dCz2RF9QVQ", "e": "AQAB", "kty": "RSA", "kid": "xrELM5bzoeSWozMrhtcrLKnmx3ZGGjycukSee3_zyks"}]}',
                "post_logout_redirect_uris": [
                    "https://client.trusted.test/disconnected"
                ],
            },
            {
                "id": "83eff494-ea5d-490e-b044-c5bcef518979",
                "created": "2025-03-15T12:55:32+00:00",
                "last_modified": "2025-03-15T12:55:32+00:00",
                "audience": [
                    "83eff494-ea5d-490e-b044-c5bcef518979",
                    "45fb7f55-816a-41bf-a9f0-16bb1d60f448",
                ],
                "client_id": "M2yzGYiZgilOvQdc8vCLyzrb",
                "client_secret": "tx22iqvdPpUBA4o8fFdGUc5jxR4596bKcs8ZvSAcLuU3LxvP",
                "client_id_issued_at": "2025-03-15T12:55:32.747508+00:00",
                "redirect_uris": [
                    "https://client.test/redirect1",
                    "https://client.test/redirect2",
                ],
                "token_endpoint_auth_method": "client_secret_basic",
                "grant_types": [
                    "password",
                    "authorization_code",
                    "implicit",
                    "hybrid",
                    "refresh_token",
                    "client_credentials",
                    "urn:ietf:params:oauth:grant-type:jwt-bearer",
                ],
                "response_types": ["code", "token", "id_token"],
                "client_name": "Some client",
                "client_uri": "https://client.test",
                "logo_uri": "https://client.test/logo.webp",
                "scope": ["openid", "email", "profile", "groups", "address", "phone"],
                "contacts": ["contact@mydomain.test"],
                "tos_uri": "https://client.test/tos",
                "policy_uri": "https://client.test/policy",
                "jwks": '{"keys": [{"n": "02uEn7s1vpHLEwNRV-5mQn0W-N--KhPoFuJ28NoRuGGBVoFmCLxnd70qdqu_8tQRuu6R65566PGiDLavkbNqe0cFZB7VSk2vnhpfJGLzUo_FYQeSK-XafnfIiuhd1IdCG6-ck7YfKJMnfqpQW_hFnq-k1SXkZVhcEsYtZ-wtXfa3hiTPrT4oFzEf0gAAvluaTTKOXaGELo-8NDv4HycQPzWLSvpAqtAEyVo_pjfAUrdDTaNPtiQG1-ISrx3UKxji0Wmt8oaJyOABQ0y3dpvT5EyvWfkooOjlebK4dI-a8ZmSU6ev_x3-HWy-9W3wkWiBCscBgiR_vmd1dCz2RF9QVQ", "e": "AQAB", "kty": "RSA", "kid": "xrELM5bzoeSWozMrhtcrLKnmx3ZGGjycukSee3_zyks"}]}',
                "post_logout_redirect_uris": ["https://client.test/disconnected"],
            },
        ],
        "consent": [
            {
                "id": "352d714d-b563-4934-8f74-ef5bba5f4cbd",
                "created": "2025-03-15T12:55:32+00:00",
                "last_modified": "2025-03-15T12:55:32+00:00",
                "consent_id": "9b97bf4c-7689-4ea2-b155-92006d192aa5",
                "subject": "17d7c987-fce0-41c3-ab48-d4de697d28e6",
                "client": "83eff494-ea5d-490e-b044-c5bcef518979",
                "scope": ["openid", "profile"],
                "issue_date": "2025-03-15T12:55:32.749027+00:00",
            }
        ],
        "token": [
            {
                "id": "57834001-55c8-4177-b22b-cbe5e62b7eb8",
                "created": "2025-03-15T12:55:32+00:00",
                "last_modified": "2025-03-15T12:55:32+00:00",
                "token_id": "iS5JFG7Yk7d4kFRUpsMNf85I0N6aCKqvs5fVwjBSKjvqafLZ",
                "access_token": "zOG6d15hkCRSzLfYlDlBvAmZ2HUeOC4FwGZLUCjVsLXE811v",
                "client": "83eff494-ea5d-490e-b044-c5bcef518979",
                "subject": "17d7c987-fce0-41c3-ab48-d4de697d28e6",
                "refresh_token": "2ilgM6qhsyWWWExKXShS9R0Chjjf2ouI72NZYIoswaXWc7iF",
                "scope": ["openid", "profile"],
                "issue_date": "2025-03-15T12:55:32.749713+00:00",
                "lifetime": 3600,
                "audience": ["83eff494-ea5d-490e-b044-c5bcef518979"],
            }
        ],
        "user": [
            {
                "id": "17d7c987-fce0-41c3-ab48-d4de697d28e6",
                "created": "2025-03-15T12:55:32+00:00",
                "last_modified": "2025-03-15T12:55:32+00:00",
                "user_name": "user",
                "password": "correct horse battery staple",
                "preferred_language": "en",
                "family_name": "Doe",
                "given_name": "John",
                "formatted_name": "John (johnny) Doe",
                "display_name": "Johnny",
                "emails": ["john@doe.test"],
                "phone_numbers": ["555-000-000"],
                "formatted_address": "1235, somewhere",
                "profile_url": "https://john.test",
            }
        ],
    }

    res = cli_runner.invoke(cli, ["restore"], input=json.dumps(payload))
    assert res.exit_code == 0, res.stdout

    client1 = backend.get(models.Client, client_id="M2yzGYiZgilOvQdc8vCLyzrb")
    client2 = backend.get(models.Client, client_id="dx8HWLJOM2GRQtkR92fjwHZ7")
    token = backend.get(models.Token)
    consent = backend.get(models.Consent)
    user = backend.get(models.User)
    assert token.client == client1
    assert consent.subject == user

    backend.delete(token)
    backend.delete(consent)
    backend.delete(client1)
    backend.delete(client2)
    backend.delete(user)



def test_incomplete_requests(testclient, logged_user, client):
    testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
        ),
        status=400,
    )


def test_bad_client(testclient, logged_user, client):
    testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            clienrt_id="nope",
        ),
        status=400,
    )



from .test_auth_otp import generate_otp


def test_signin_with_multiple_otp_methods(smtpd, testclient, backend, user):
    testclient.app.config["CANAILLE"]["AUTHENTICATION_FACTORS"] = [
        "password",
        "otp",
        "email",
        "sms",
    ]

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    # TOTP/HOTP
    res.form["otp"] = generate_otp("TOTP", user.secret_token)
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    # EMAIL_OTP
    backend.reload(user)
    res.form["otp"] = user.one_time_password
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    # SMS_OTP
    backend.reload(user)
    res.form["otp"] = user.one_time_password
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes


def test_login_page_resets_multifactor_authentication_progress(testclient, user):
    """Verify visiting login page resets any in-progress multifactor authentication session."""
    testclient.app.config["CANAILLE"]["AUTHENTICATION_FACTORS"] = ["password", "otp"]

    # Start authentication process with first user
    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res = res.form.submit(status=302).follow(status=200)

    # Complete password step
    assert res.template == "core/auth/password.html"
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302).follow(status=200)

    # Should be on OTP page (second factor)
    assert res.template == "core/auth/otp.html"

    # Return to login page directly (simulating user changing mind)
    res = testclient.get("/login")

    # Submit a different username
    res.form["login"] = "differentuser"
    res = res.form.submit(status=302).follow(status=200)

    # Should be back on password page (first step) - authentication was reset
    assert res.template == "core/auth/password.html"

    # Verify we're authenticating the new user
    assert "differentuser" in res.text



import typing

import canaille.core.models
import canaille.oidc.models
from canaille.app import models
from canaille.backends import Backend
from canaille.backends.models import BackendModel


class MemoryModel(BackendModel):
    def __init__(self, *args, **kwargs):
        self._state = {}
        self._cache = {}
        for attribute, value in kwargs.items():
            setattr(self, attribute, value)

    def __repr__(self):
        return f"<{self.__class__.__name__} id={self.id}>"

    @classmethod
    def serialize(cls, value):
        if isinstance(value, list):
            values = [cls.serialize(item) for item in value]
            return [item for item in values if item is not None]

        return value.id if isinstance(value, MemoryModel) else value

    @classmethod
    def deserialize(cls, attribute_name, value):
        if isinstance(value, list):
            values = [cls.deserialize(attribute_name, item) for item in value]
            return [item for item in values if item is not None]

        if value is None or value == "":
            multiple_attribute = (
                typing.get_origin(cls.attributes[attribute_name]) is list
            )
            return [] if multiple_attribute else None

        model, _ = cls.get_model_annotations(attribute_name)
        if model and not isinstance(value, model):
            backend_model = getattr(models, model.__name__)
            return Backend.instance.get(backend_model, id=value)

        return value

    def __eq__(self, other):
        if other is None:
            return False

        if not isinstance(other, MemoryModel):
            return self == Backend.instance.get(self.__class__, id=other)

        return self._state == other._state

    def __hash__(self):
        return hash(self.id)

    def __getattribute__(self, name):
        if name != "attributes" and name in self.attributes:
            return self.deserialize(name, self._cache.get(name, self._state.get(name)))

        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if name in self.attributes:
            self._cache[name] = value
            self._state[name] = self.serialize(value)
        else:
            super().__setattr__(name, value)


class User(canaille.core.models.User, MemoryModel):
    def get_password_hash(self):
        return self.password


class Group(canaille.core.models.Group, MemoryModel):
    pass


class Client(canaille.oidc.models.Client, MemoryModel):
    pass


class AuthorizationCode(canaille.oidc.models.AuthorizationCode, MemoryModel):
    pass


class Token(canaille.oidc.models.Token, MemoryModel):
    pass


class Consent(canaille.oidc.models.Consent, MemoryModel):
    pass



import unittest.mock

from canaille.app import models
from canaille.backends import Backend
from canaille.commands import cli


def test_delete_all_with_confirmation(cli_runner, backend):
    """Remove all models by default."""
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)

    res = cli_runner.invoke(cli, ["delete", "user"], input="Y")
    assert res.exit_code == 0, res.stdout
    assert "1 item(s) deleted." in res.stdout.splitlines()

    assert not backend.get(models.User, user_name="foobar")


def test_delete_all_without_confirmation(cli_runner, backend):
    """Refuse to remove all models by default."""
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)

    res = cli_runner.invoke(cli, ["delete", "user"], input="N")
    assert res.exit_code == 0, res.stdout
    assert "1 item(s) deleted." not in res.stdout.splitlines()

    assert backend.get(models.User, user_name="foobar")
    backend.delete(user)


def test_delete_all_without_confirmation_prompt(cli_runner, backend):
    """Remove all models by default."""
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)

    res = cli_runner.invoke(cli, ["delete", "user", "--noconfirm"])
    assert res.exit_code == 0, res.stdout
    assert "1 item(s) deleted." in res.stdout.splitlines()

    assert not backend.get(models.User, user_name="foobar")


def test_delete_by_id(cli_runner, backend):
    """Remove a model identified by its id."""
    user = models.User(
        formatted_name="Foo bar",
        family_name="Bar",
        emails=["foobar@example.test"],
        user_name="foobar",
    )
    backend.save(user)

    res = cli_runner.invoke(cli, ["delete", "user", "--id", user.id, "--noconfirm"])
    assert res.exit_code == 0, res.stdout
    assert "1 item(s) deleted." in res.stdout.splitlines()

    assert not backend.get(models.User, user_name="foobar")


def test_delete_unknown_id(cli_runner, backend):
    """Error case for trying to set a value for an invalid object."""
    res = cli_runner.invoke(cli, ["delete", "user", "--id", "invalid", "--noconfirm"])
    assert res.exit_code == 0, res.stdout
    assert "0 item(s) deleted." in res.stdout.splitlines()


def test_delete_ignore_errors_query_failure(cli_runner, backend):
    """Test that query errors are ignored with --ignore-errors."""

    def mock_query(*args, **kwargs):
        raise Exception("Simulated query error")

    with unittest.mock.patch.object(Backend.instance, "query", side_effect=mock_query):
        res = cli_runner.invoke(cli, ["delete", "user", "--noconfirm"])
        assert res.exit_code != 0
        assert "Simulated query error" in str(res.output)

    with unittest.mock.patch.object(Backend.instance, "query", side_effect=mock_query):
        res = cli_runner.invoke(
            cli, ["delete", "user", "--noconfirm", "--ignore-errors"]
        )
        assert res.exit_code == 0
        assert res.stdout == ""


def test_delete_ignore_errors_partial_failure(cli_runner, backend):
    """Test that --ignore-errors skips items that fail to delete."""
    users = []
    for i in range(3):
        user = models.User(
            formatted_name=f"Test User {i}",
            family_name=f"TestUser{i}",
            emails=[f"testuser{i}@example.test"],
            user_name=f"testuser{i}",
        )
        backend.save(user)
        users.append(user)

    original_delete = Backend.instance.delete

    def mock_delete(obj):
        if obj.user_name == "testuser1":
            raise Exception("Cannot delete testuser1")
        return original_delete(obj)

    with unittest.mock.patch.object(
        Backend.instance, "delete", side_effect=mock_delete
    ):
        res = cli_runner.invoke(cli, ["delete", "user", "--noconfirm"])
        assert res.exit_code != 0
        assert "Cannot delete testuser1" in str(res.output)

    total_users = len(backend.query(models.User))
    expected_deleted = total_users - 1

    with unittest.mock.patch.object(
        Backend.instance, "delete", side_effect=mock_delete
    ):
        res = cli_runner.invoke(
            cli, ["delete", "user", "--noconfirm", "--ignore-errors"]
        )
        assert res.exit_code == 0
        assert f"{expected_deleted} item(s) deleted." in res.stdout

    remaining = backend.get(models.User, user_name="testuser1")
    backend.delete(remaining)



import datetime
import json

import httpx
from authlib.integrations.flask_oauth2 import AuthorizationServer
from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.oauth2 import rfc6749
from authlib.oauth2 import rfc6750
from authlib.oauth2 import rfc7009
from authlib.oauth2 import rfc7523
from authlib.oauth2 import rfc7591
from authlib.oauth2 import rfc7592
from authlib.oauth2 import rfc7636
from authlib.oauth2 import rfc7662
from authlib.oauth2 import rfc8414
from authlib.oauth2 import rfc9101
from authlib.oauth2 import rfc9207
from authlib.oauth2.rfc6749 import InvalidClientError
from authlib.oidc import core as oidc_core
from authlib.oidc import discovery as oidc_discovery
from authlib.oidc import registration as oidc_registration
from authlib.oidc.core.grants.util import generate_id_token
from flask import current_app
from flask import g
from flask import request
from flask import url_for
from joserfc import jwk
from joserfc import jws
from werkzeug.security import gen_salt

from canaille.app import DOCUMENTATION_URL
from canaille.app import models
from canaille.app.flask import cache
from canaille.backends import Backend
from canaille.core.auth import get_user_from_login

from .jwk import make_default_jwk

AUTHORIZATION_CODE_LIFETIME = 84400
JWT_JTI_CACHE_LIFETIME = 3600

registry = jws.JWSRegistry(algorithms=list(jws.JWSRegistry.algorithms.keys()))


class AuthorizationServerMetadata(
    rfc8414.AuthorizationServerMetadata, rfc9101.AuthorizationServerMetadata
):
    pass


def oauth_authorization_server():
    payload = {
        "issuer": get_issuer(),
        "authorization_endpoint": url_for("oidc.endpoints.authorize", _external=True),
        "token_endpoint": url_for("oidc.endpoints.issue_token", _external=True),
        "token_endpoint_auth_methods_supported": [
            "client_secret_basic",
            "private_key_jwt",
            "client_secret_post",
            "none",
        ],
        "token_endpoint_auth_signing_alg_values_supported": [
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "userinfo_endpoint": url_for("oidc.endpoints.userinfo", _external=True),
        "revocation_endpoint": url_for("oidc.endpoints.revoke_token", _external=True),
        "revocation_endpoint_auth_methods_supported": ["client_secret_basic"],
        "revocation_endpoint_auth_signing_alg_values_supported": None,
        "introspection_endpoint": url_for(
            "oidc.endpoints.introspect_token", _external=True
        ),
        "introspection_endpoint_auth_methods_supported": ["client_secret_basic"],
        "introspection_endpoint_auth_signing_alg_values_supported": None,
        "jwks_uri": url_for("oidc.endpoints.jwks", _external=True),
        "registration_endpoint": url_for(
            "oidc.endpoints.client_registration", _external=True
        ),
        "scopes_supported": [
            "openid",
            "profile",
            "email",
            "address",
            "phone",
            "groups",
        ],
        "response_types_supported": [
            "code",
            "token",
            "id_token",
            "code token",
            "code id_token",
            "token id_token",
            "code id_token token",
        ],
        "grant_types_supported": [
            "authorization_code",
            "implicit",
            "password",
            "client_credentials",
            "refresh_token",
        ],
        "ui_locales_supported": g.available_language_codes,
        "code_challenge_methods_supported": CodeChallenge.SUPPORTED_CODE_CHALLENGE_METHOD,
        "service_documentation": DOCUMENTATION_URL,
        "authorization_response_iss_parameter_supported": True,
        "op_policy_uri": None,
        "op_tos_uri": None,
    }
    obj = AuthorizationServerMetadata(payload)
    obj.validate()
    return obj


def openid_configuration():
    prompt_values_supported = ["none", "login", "consent"] + (
        ["create"] if current_app.config["CANAILLE"]["ENABLE_REGISTRATION"] else []
    )
    payload = {
        **oauth_authorization_server(),
        "end_session_endpoint": url_for("oidc.endpoints.end_session", _external=True),
        "claim_types_supported": ["normal"],
        "claims_supported": [
            "sub",
            "iss",
            "auth_time",
            "acr",
            "name",
            "given_name",
            "family_name",
            "nickname",
            "profile",
            "picture",
            "website",
            "email",
            "email_verified",
            "locale",
            "zoneinfo",
            "groups",
            "nonce",
        ],
        "claims_locales_supported": None,
        "claims_parameter_supported": False,
        "subject_types_supported": ["public"],
        "id_token_signing_alg_values_supported": [
            "none",
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "id_token_encryption_alg_values_supported": None,
        "id_token_encryption_enc_values_supported": None,
        "userinfo_signing_alg_values_supported": [
            "none",
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "userinfo_encryption_alg_values_supported": None,
        "userinfo_encryption_enc_values_supported": None,
        "prompt_values_supported": prompt_values_supported,
        "request_parameter_supported": True,
        "request_uri_parameter_supported": True,
        "require_request_uri_registration": False,
        "request_object_signing_alg_values_supported": [
            "none",
            "RS256",
            "ES256",
            "HS256",
            "EdDSA",
        ],
        "request_object_encryption_alg_values_supported": None,
        "request_object_encryption_enc_values_supported": None,
        "response_modes_supported": ["query", "fragment"],
        "acr_values_supported": None,
        "display_values_supported": None,
    }
    obj = oidc_discovery.OpenIDProviderMetadata(payload)
    obj.validate()
    return obj


def exists_nonce(nonce, request):
    client = Backend.instance.get(models.Client, id=request.payload.client_id)
    exists = Backend.instance.query(
        models.AuthorizationCode, client=client, nonce=nonce
    )
    return bool(exists)


def get_alg_for_key(key):
    """Find the algorithm for the given key."""
    # TODO: Improve this when a better solution is implemented upstream
    # https://github.com/authlib/joserfc/issues/49
    for alg_name, alg in registry.algorithms.items():  # pragma: no cover
        if alg.key_type == key.key_type:
            return alg_name


def get_issuer():
    if server_name := current_app.config.get("SERVER_NAME"):
        scheme = current_app.config.get("PREFERRED_URL_SCHEME", "http")
        return f"{scheme}://{server_name}"

    return request.url_root


def server_jwks(include_inactive=True):
    keys = list(current_app.config["CANAILLE_OIDC"]["ACTIVE_JWKS"])
    if include_inactive and current_app.config["CANAILLE_OIDC"]["INACTIVE_JWKS"]:
        keys += list(current_app.config["CANAILLE_OIDC"]["INACTIVE_JWKS"])

    key_objs = [jwk.import_key(key) for key in keys if key]
    for obj in key_objs:
        obj.ensure_kid()
    return jwk.KeySet(key_objs)


def get_jwt_config(grant=None):
    jwks = server_jwks(include_inactive=False)
    jwk = jwks.keys[0]
    payload = {
        "key": jwk.as_dict(),
        "iss": get_issuer(),
        "kid": jwk.kid,
        "alg": get_alg_for_key(jwk),
    }
    return payload


def get_client_jwks(client, kid=None):
    """Get the client JWK set, either stored locally or by downloading them from the URI the client indicated."""

    @cache.cached(timeout=50, key_prefix=f"jwks_{client.client_id}")
    def get_public_jwks():
        return httpx.get(client.jwks_uri).json()

    if client.jwks_uri:
        raw_jwks = get_public_jwks()
        key_set = jwk.KeySet.import_key_set(raw_jwks)
        key = key_set.get_by_kid(kid)
        return key

    if client.jwks:
        raw_jwks = json.loads(client.jwks)
        key_set = jwk.KeySet.import_key_set(raw_jwks)
        key = key_set.get_by_kid(kid)
        return key

    return None


class UserInfo(oidc_core.UserInfo):
    REGISTERED_CLAIMS = oidc_core.UserInfo.REGISTERED_CLAIMS + ["groups"]
    SCOPES_CLAIMS_MAPPING = {
        "groups": ["groups"],
        **oidc_core.UserInfo.SCOPES_CLAIMS_MAPPING,
    }


def make_address_claim(user):
    payload = {}
    mapping = {
        "formatted_address": "formatted",
        "street": "street_address",
        "locality": "locality",
        "region": "region",
        "postal_code": "postal_code",
    }
    for user_attr, claim in mapping.items():
        if val := getattr(user, user_attr):
            payload[claim] = val

    return payload


def generate_user_claims(user, jwt_mapping_config=None):
    jwt_mapping_config = {
        **(current_app.config["CANAILLE_OIDC"]["USERINFO_MAPPING"]),
        **(jwt_mapping_config or {}),
    }

    data = {}
    for claim in UserInfo.REGISTERED_CLAIMS:
        raw_claim = jwt_mapping_config.get(claim.upper())
        if raw_claim:
            formatted_claim = current_app.jinja_env.from_string(raw_claim).render(
                user=user
            )
            if formatted_claim:
                # According to https://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse
                # it's better to not insert a null or empty string value
                data[claim] = formatted_claim

        if claim == "address" and (address := make_address_claim(user)):
            data[claim] = address

        if claim == "groups" and user.groups:
            data[claim] = [group.display_name for group in user.groups]

        if claim == "updated_at":
            data[claim] = int(float(data[claim]))

    return data


def save_authorization_code(code, request):
    nonce = request.payload.data.get("nonce")
    now = datetime.datetime.now(datetime.timezone.utc)
    scope = request.client.get_allowed_scope(request.payload.scope)
    code = models.AuthorizationCode(
        authorization_code_id=gen_salt(48),
        code=code,
        subject=request.user,
        client=request.client,
        redirect_uri=request.payload.redirect_uri or request.client.redirect_uris[0],
        scope=scope.split(" "),
        nonce=nonce,
        issue_date=now,
        lifetime=AUTHORIZATION_CODE_LIFETIME,
        challenge=request.payload.data.get("code_challenge"),
        challenge_method=request.payload.data.get("code_challenge_method"),
        auth_time=g.session.last_login_datetime,
    )
    Backend.instance.save(code)
    return code.code


class JWTClientAuth(rfc7523.JWTBearerClientAssertion):
    def validate_jti(self, claims, jti):
        """Indicate whether the jti was used before."""
        key = "jti:{}-{}".format(claims["sub"], jti)
        if cache.get(key):
            return False
        cache.set(key, 1, timeout=JWT_JTI_CACHE_LIFETIME)
        return True

    def resolve_client_public_key(self, client, headers):
        jwk = get_client_jwks(client)
        if not jwk:
            raise InvalidClientError(description="No matching JWK")

        return jwk.as_dict()


class AuthorizationCodeGrant(rfc6749.AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = [
        "client_secret_basic",
        "client_secret_post",
        JWTClientAuth.CLIENT_AUTH_METHOD,
        "none",
    ]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)

    def query_authorization_code(self, code, client):
        item = Backend.instance.query(
            models.AuthorizationCode, code=code, client=client
        )
        if item and not item[0].is_expired():
            return item[0]

    def delete_authorization_code(self, authorization_code):
        Backend.instance.delete(authorization_code)

    def authenticate_user(self, authorization_code):
        if authorization_code.subject and not authorization_code.subject.locked:
            return authorization_code.subject


class OpenIDCode(oidc_core.OpenIDCode):
    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def get_jwt_config(self, grant=None):
        # Hotfix until fixed upstream:
        # client.id_token_signed_response_alg should be used when defined,
        # this can only happen if 'alg' is not set
        # https://github.com/authlib/authlib/issues/806
        result = get_jwt_config(grant)
        del result["alg"]
        return result

    def generate_user_info(self, user, scope):
        return UserInfo(generate_user_claims(user)).filter(scope)

    def get_audiences(self, request):
        client = request.client
        return [aud.client_id for aud in client.audience]


class PasswordGrant(rfc6749.ResourceOwnerPasswordCredentialsGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = [
        "client_secret_basic",
        "client_secret_post",
        JWTClientAuth.CLIENT_AUTH_METHOD,
        "none",
    ]

    def authenticate_user(self, username, password):
        user = get_user_from_login(username)
        if not user:
            return None

        success, _ = Backend.instance.check_user_password(user, password)
        if not success:
            return None

        return user


class RefreshTokenGrant(rfc6749.RefreshTokenGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = [
        "client_secret_basic",
        "client_secret_post",
        JWTClientAuth.CLIENT_AUTH_METHOD,
        "none",
    ]

    def authenticate_refresh_token(self, refresh_token):
        token = Backend.instance.query(models.Token, refresh_token=refresh_token)
        if token and token[0].is_refresh_token_active():
            return token[0]

    def authenticate_user(self, credential):
        if credential.subject and not credential.subject.locked:
            return credential.subject

    def revoke_old_credential(self, credential):
        credential.revokation_date = datetime.datetime.now(datetime.timezone.utc)
        Backend.instance.save(credential)


class JWTBearerGrant(rfc7523.JWTBearerGrant):
    def resolve_issuer_client(self, issuer):
        return Backend.instance.get(models.Client, client_id=issuer)

    def resolve_client_key(self, client, headers, payload):
        jwk = server_jwks().keys[0]
        return jwk.as_dict()

    def authenticate_user(self, subject: str):
        return get_user_from_login(subject)

    def has_granted_permission(self, client, user):
        grant = Backend.instance.get(models.Consent, client=client, subject=user)
        has_permission = (grant and not grant.revoked) or (not grant and client.trusted)
        return has_permission


class OpenIDImplicitGrant(oidc_core.OpenIDImplicitGrant):
    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def get_jwt_config(self, grant=None):
        # Hotfix until fixed upstream:
        # client.id_token_signed_response_alg should be used when defined,
        # this can only happen if 'alg' is not set
        # https://github.com/authlib/authlib/issues/806
        result = get_jwt_config(grant)
        del result["alg"]
        return result

    def generate_user_info(self, user, scope):
        return UserInfo(generate_user_claims(user)).filter(scope)

    def get_audiences(self, request):
        client = request.client
        return [aud.client_id for aud in client.audience]


class OpenIDHybridGrant(oidc_core.OpenIDHybridGrant):
    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)

    def exists_nonce(self, nonce, request):
        return exists_nonce(nonce, request)

    def get_jwt_config(self, grant=None):
        # Hotfix until fixed upstream:
        # client.id_token_signed_response_alg should be used when defined,
        # this can only happen if 'alg' is not set
        # https://github.com/authlib/authlib/issues/806
        result = get_jwt_config(grant)
        del result["alg"]
        return result

    def generate_user_info(self, user, scope):
        return UserInfo(generate_user_claims(user)).filter(scope)

    def get_audiences(self, request):
        client = request.client
        return [aud.client_id for aud in client.audience]


def query_client(client_id):
    return Backend.instance.get(models.Client, client_id=client_id)


def save_token(token, request):
    now = datetime.datetime.now(datetime.timezone.utc)
    t = models.Token(
        token_id=gen_salt(48),
        type=token["token_type"],
        access_token=token["access_token"],
        issue_date=now,
        lifetime=token["expires_in"],
        scope=token["scope"].split(" "),
        client=request.client,
        refresh_token=token.get("refresh_token"),
        subject=request.user,
        audience=request.client.audience,
    )
    Backend.instance.save(t)


class BearerTokenValidator(rfc6750.BearerTokenValidator):
    def authenticate_token(self, token_string):
        return Backend.instance.get(models.Token, access_token=token_string)


def query_token(token, token_type_hint):
    if token_type_hint == "access_token":
        return Backend.instance.get(models.Token, access_token=token)
    elif token_type_hint == "refresh_token":
        return Backend.instance.get(models.Token, refresh_token=token)

    item = Backend.instance.get(models.Token, access_token=token)
    if item:
        return item

    item = Backend.instance.get(models.Token, refresh_token=token)
    if item:
        return item

    return None


class RevocationEndpoint(rfc7009.RevocationEndpoint):
    def query_token(self, token, token_type_hint):
        return query_token(token, token_type_hint)

    def revoke_token(self, token, request):
        token.revokation_date = datetime.datetime.now(datetime.timezone.utc)
        Backend.instance.save(token)


class IntrospectionEndpoint(rfc7662.IntrospectionEndpoint):
    def query_token(self, token, token_type_hint):
        return query_token(token, token_type_hint)

    def check_permission(self, token, client, request):
        return client in token.audience

    def introspect_token(self, token):
        audience = [aud.client_id for aud in token.audience]
        response = {
            "active": True,
            "client_id": token.client.client_id,
            "token_type": token.type,
            "scope": token.get_scope(),
            "aud": audience,
            "iss": get_issuer(),
            "exp": token.get_expires_at(),
            "iat": token.get_issued_at(),
        }
        if token.subject:
            response["username"] = token.subject.formatted_name
            response["sub"] = token.subject.user_name
        return response


class ClientManagementMixin:
    def authenticate_token(self, request):
        if current_app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_OPEN"]:
            return True

        auth_header = request.headers.get("Authorization")
        if not auth_header or not auth_header.lower().startswith("bearer "):
            return None

        bearer_token = auth_header.split()[1]
        if bearer_token not in (
            current_app.config["CANAILLE_OIDC"]["DYNAMIC_CLIENT_REGISTRATION_TOKENS"]
            or []
        ):
            return None

        return True

    def get_server_metadata(self):
        result = openid_configuration()
        return result

    def resolve_public_key(self, request):
        return server_jwks().keys[0].as_dict(private=False)

    def client_convert_data(self, **kwargs):
        if "client_id_issued_at" in kwargs:
            kwargs["client_id_issued_at"] = datetime.datetime.fromtimestamp(
                kwargs["client_id_issued_at"], datetime.timezone.utc
            )

        if "client_secret_expires_at" in kwargs:
            kwargs["client_secret_expires_at"] = datetime.datetime.fromtimestamp(
                kwargs["client_secret_expires_at"], datetime.timezone.utc
            )

        if "jwks" in kwargs:
            kwargs["jwks"] = json.dumps(kwargs["jwks"])

        if "scope" in kwargs:
            kwargs["scope"] = kwargs["scope"].split(" ")

        if "response_types" in kwargs:
            kwargs["response_types"] = list(
                {
                    item
                    for response_type in kwargs["response_types"]
                    for item in response_type.split(" ")
                }
            )

        return kwargs


class ClientRegistrationEndpoint(
    ClientManagementMixin, rfc7591.ClientRegistrationEndpoint
):
    software_statement_alg_values_supported = [
        "none",
        "RS256",
        "ES256",
        "HS256",
        "EdDSA",
    ]

    def generate_client_registration_info(self, client, request):
        payload = {
            "scope": " ".join(client.scope) if client.scope else "",
            "registration_client_uri": url_for(
                "oidc.endpoints.client_registration_management",
                client_id=client.client_id,
                _external=True,
            ),
        }

        if "Authorization" in request.headers:
            access_token = request.headers["Authorization"].split(" ")[1]
            payload["registration_access_token"] = access_token

        return payload

    def save_client(self, client_info, client_metadata, request):
        client = models.Client(
            # this won't be needed when OIDC RP Initiated Logout is
            # directly implemented in authlib:
            # https://gitlab.com/yaal/canaille/-/issues/157
            post_logout_redirect_uris=request.payload.data.get(
                "post_logout_redirect_uris"
            ),
            **self.client_convert_data(**client_info, **client_metadata),
        )
        Backend.instance.save(client)
        client.audience = [client]
        Backend.instance.save(client)
        return client


class ClientConfigurationEndpoint(
    ClientManagementMixin, rfc7592.ClientConfigurationEndpoint
):
    def authenticate_client(self, request):
        client_id = request.uri.split("/")[-1]
        return Backend.instance.get(models.Client, client_id=client_id)

    def revoke_access_token(self, request, token):
        pass

    def check_permission(self, client, request):
        return True

    def delete_client(self, client, request):
        Backend.instance.delete(client)

    def update_client(self, client, client_metadata, request):
        Backend.instance.update(client, **self.client_convert_data(**client_metadata))
        Backend.instance.save(client)
        return client

    def generate_client_registration_info(self, client, request):
        access_token = request.headers["Authorization"].split(" ")[1]
        return {
            "registration_client_uri": request.uri,
            "registration_access_token": access_token,
        }


class CodeChallenge(rfc7636.CodeChallenge):
    def get_authorization_code_challenge(self, authorization_code):
        return authorization_code.challenge

    def get_authorization_code_challenge_method(self, authorization_code):
        return authorization_code.challenge_method


class UserInfoEndpoint(oidc_core.UserInfoEndpoint):
    def get_issuer(self):
        return get_issuer()

    def generate_user_info(self, user, scope):
        return UserInfo(generate_user_claims(user)).filter(scope)

    def resolve_private_key(self):
        return server_jwks(include_inactive=False).as_dict()


class IssuerParameter(rfc9207.IssuerParameter):
    def get_issuer(self) -> str:
        return get_issuer()


class JWTAuthenticationRequest(rfc9101.JWTAuthenticationRequest):
    def resolve_client_public_key(self, client):
        return get_client_jwks(client).as_dict()

    def get_request_object(self, request_uri: str):
        return httpx.get(request_uri).text

    def get_server_metadata(self):
        return openid_configuration()

    def get_client_require_signed_request_object(self, client):
        return client.client_metadata.get("require_signed_request_object", False)


authorization = AuthorizationServer()
require_oauth = ResourceProtector()


def generate_access_token(client, grant_type, user, scope):
    if grant_type == "client_credentials":
        # Canaille could generate a JWT with iss/sub/aud/exp/iat/jti/scope/client_id
        # instead of a random string
        return gen_salt(48)

    audience = [client.client_id for client in client.audience]
    bearer_token_generator = authorization._token_generators["default"]
    kwargs = {
        "token": {},
        "user_info": UserInfo(generate_user_claims(user)).filter(scope),
        "aud": audience,
        **get_jwt_config(grant_type),
    }
    kwargs["exp"] = bearer_token_generator._get_expires_in(client, grant_type)
    return generate_id_token(**kwargs)


def setup_oauth(app):
    app.config["OAUTH2_REFRESH_TOKEN_GENERATOR"] = True
    app.config["OAUTH2_ACCESS_TOKEN_GENERATOR"] = (
        "canaille.oidc.oauth.generate_access_token"
    )

    oidc_config = app.config.get("CANAILLE_OIDC")
    if oidc_config and not oidc_config.get("ACTIVE_JWKS"):
        oidc_config["ACTIVE_JWKS"] = [
            make_default_jwk(app.config.get("SECRET_KEY")).as_dict()
        ]

    # hacky, but needed for tests as somehow the same 'authorization' object is used
    # between tests
    authorization.__init__()
    authorization.init_app(app, query_client=query_client, save_token=save_token)

    authorization.register_grant(rfc6749.ClientCredentialsGrant)
    authorization.register_grant(rfc6749.ImplicitGrant)
    authorization.register_grant(PasswordGrant)
    authorization.register_grant(RefreshTokenGrant)
    authorization.register_grant(JWTBearerGrant)
    authorization.register_grant(OpenIDImplicitGrant)
    authorization.register_grant(OpenIDHybridGrant)
    authorization.register_grant(
        AuthorizationCodeGrant,
        [
            OpenIDCode(require_nonce=app.config["CANAILLE_OIDC"]["REQUIRE_NONCE"]),
            CodeChallenge(required=True),
        ],
    )

    with app.app_context():
        if not app.config["SERVER_NAME"]:
            app.logger.warning(
                "The 'SERVER_NAME' configuration parameter is unset. JWT client authentication is disabled."
            )

        else:
            authorization.register_client_auth_method(
                JWTClientAuth.CLIENT_AUTH_METHOD,
                JWTClientAuth(url_for("oidc.endpoints.issue_token", _external=True)),
            )

    require_oauth.register_token_validator(BearerTokenValidator())

    authorization.register_endpoint(UserInfoEndpoint(resource_protector=require_oauth))
    authorization.register_endpoint(IntrospectionEndpoint)
    authorization.register_endpoint(RevocationEndpoint)
    authorization.register_endpoint(
        ClientRegistrationEndpoint(
            claims_classes=[
                rfc7591.ClientMetadataClaims,
                rfc9101.ClientMetadataClaims,
                oidc_registration.ClientMetadataClaims,
            ]
        )
    )
    authorization.register_endpoint(
        ClientConfigurationEndpoint(
            claims_classes=[
                rfc7591.ClientMetadataClaims,
                rfc9101.ClientMetadataClaims,
                oidc_registration.ClientMetadataClaims,
            ]
        )
    )

    authorization.register_extension(IssuerParameter())
    authorization.register_extension(JWTAuthenticationRequest())



import datetime

from canaille.core.auth import AuthenticationSession
from canaille.core.auth import get_user_from_login


def test_user_get_user_from_login(testclient, user):
    """Check that 'get_user_from_login' gets users according to LOGIN_ATTRIBUTES."""
    assert get_user_from_login(login="invalid") is None
    assert get_user_from_login(login="user") == user
    assert get_user_from_login(login="john@doe.test") == user
    assert get_user_from_login(login="555-000-000") is None


def test_user_get_user_from_login_dict(testclient, user):
    """Check that 'get_user_from_login' gets users according to LOGIN_ATTRIBUTES when it is a dict."""
    testclient.app.config["CANAILLE"]["LOGIN_ATTRIBUTES"] = {
        "user_name": "{{ login | replace('@doe.test', '') }}",
        "emails": "{{ login }}",
    }
    assert get_user_from_login(login="invalid") is None
    assert get_user_from_login(login="user") == user
    assert get_user_from_login(login="john@doe.test") == user
    assert get_user_from_login(login="user@doe.test") == user


def test_visitor_logout(testclient, user):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/logout")
    res = res.follow(status=302)
    res = res.follow(status=200)
    assert (
        "success",
        "You have been disconnected. See you next time user",
    ) not in res.flashes

    with testclient.session_transaction() as session:
        assert not session.get("sessions")


def test_wrong_login(testclient, user):
    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = True

    res = testclient.get("/login", status=200)
    res.form["login"] = "invalid"
    res = res.form.submit(status=302)
    res = res.follow(status=200)
    assert res.template == "core/auth/password.html"

    testclient.app.config["CANAILLE"]["HIDE_INVALID_LOGINS"] = False

    res = testclient.get("/login", status=200)
    res.form["login"] = "invalid"
    res = res.form.submit(status=200)
    res.mustcontain("The login 'invalid' does not exist")


def test_signin_locked_account(testclient, user, backend):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    user.lock_date = datetime.datetime.now(datetime.timezone.utc)
    backend.save(user)

    res = testclient.get("/login", status=200)
    res.form["login"] = "user"

    res = res.form.submit(status=302)
    res = res.follow(status=200)
    res.form["password"] = "correct horse battery staple"

    res = res.form.submit()
    res.mustcontain("Your account has been locked.")

    user.lock_date = None
    backend.save(user)


def test_login_placeholder(testclient):
    """Check that the login form placeholders display values according to the 'LOGIN_ATTRIBUTES' configuration parameter."""
    testclient.app.config["CANAILLE"]["LOGIN_ATTRIBUTES"] = ["user_name"]
    placeholder = testclient.get("/login").form["login"].attrs["placeholder"]
    assert placeholder == "jdoe"

    testclient.app.config["CANAILLE"]["LOGIN_ATTRIBUTES"] = ["formatted_name"]
    placeholder = testclient.get("/login").form["login"].attrs["placeholder"]
    assert placeholder == "John Doe"

    testclient.app.config["CANAILLE"]["LOGIN_ATTRIBUTES"] = ["emails"]
    placeholder = testclient.get("/login").form["login"].attrs["placeholder"]
    assert placeholder == "john.doe@example.com"

    testclient.app.config["CANAILLE"]["LOGIN_ATTRIBUTES"] = ["user_name", "emails"]
    placeholder = testclient.get("/login").form["login"].attrs["placeholder"]
    assert placeholder == "jdoe or john.doe@example.com"


def test_bad_authentication_step(testclient, user):
    """Handle corrupted authentication sessions."""
    with testclient.session_transaction() as session:
        session["auth"] = AuthenticationSession(
            user_name="user",
            remaining=["password", "invalid"],
        ).serialize()

    res = testclient.get("/auth/password", status=200)
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)

    assert (
        "error",
        "An error happened during your authentication process. Please try again.",
    ) in res.flashes


def test_get_user_from_login_partial_email_no_multiple_results(
    testclient, user, moderator, backend
):
    """Test that partial email searches don't cause MultipleResultsFound errors (issue #278).

    This test reproduces the SQLAlchemy bug where searching for a partial string
    in JSON list columns would find multiple users and cause scalar_one_or_none()
    to raise MultipleResultsFound instead of returning None.
    """
    # Ensure we have users with emails containing common substrings
    user.emails = ["user@example.com"]
    moderator.emails = ["support@example.com"]
    backend.save(user)
    backend.save(moderator)

    result = get_user_from_login(login="u")

    assert result is None
    exact_result = get_user_from_login(login="user@example.com")
    assert exact_result == user


def test_session_permanence_with_remember_true(testclient, user, backend):
    # This test verifies the session is configured as permanent
    # Actual expiration testing would require browser simulation
    from canaille.app.session import LOGIN_HISTORY

    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get("/login")
    res.form["login"] = user.user_name
    res = res.form.submit()

    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    # Note: WebTest doesn't expose session.permanent directly
    # This test mainly ensures the login flow works with remember=True
    with testclient.session_transaction() as sess:
        assert LOGIN_HISTORY in sess
        assert user.user_name in sess[LOGIN_HISTORY]


def test_direct_login_with_username_uses_default_remember(testclient, user, backend):
    """Test that direct login via /login/<username> uses default remember=True."""
    from canaille.app.session import LOGIN_HISTORY

    with testclient.session_transaction() as sess:
        sess.clear()

    res = testclient.get(f"/login/{user.user_name}")
    res = res.follow()
    res.form["password"] = "correct horse battery staple"
    res = res.form.submit()

    with testclient.session_transaction() as sess:
        assert LOGIN_HISTORY in sess
        assert user.user_name in sess[LOGIN_HISTORY]



import os
import pathlib
import sys
from urllib.parse import urlparse
from wsgiref.simple_server import WSGIRequestHandler

import tomlkit

sys.path.insert(0, os.path.abspath(".."))
from dev.devapp import create_app

WSGIRequestHandler.log_request = lambda *args, **kwargs: None


def create_doc_app(sphinx_app):
    conf_path = pathlib.Path(__file__).parent.parent / "dev" / "conf" / "canaille.toml"
    with open(conf_path) as fd:
        conf = dict(tomlkit.load(fd))
    conf["CANAILLE"]["DATABASE"] = "memory"
    conf["CANAILLE"]["SECRET_KEY"] = "doc"
    conf["CANAILLE"]["SMTP"] = {"HOST": "localhost"}
    conf["CANAILLE"]["LANGUAGE"] = sphinx_app.config["language"]
    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"
    app = create_app(conf)
    return app


def get_root_url(url):
    parsed_url = urlparse(url)
    root_url = f"{parsed_url.scheme}://{parsed_url.hostname}"
    if parsed_url.port:
        root_url += f":{parsed_url.port}"
    return root_url


contexts = {}


def context_login(browser, url, color_scheme, user, password=True):
    """Create a playwright context where 'user' has been authenticated.

    It helps playwright take screenshot as a logged in user.
    """
    try:
        context = browser.new_context(
            color_scheme=color_scheme, storage_state=contexts[user]
        )
    except KeyError:
        context = browser.new_context(color_scheme=color_scheme)

        page = context.new_page()
        page.goto(get_root_url(url) + "/login")
        page.wait_for_load_state()
        page.locator("input[name=login]").fill(user)
        page.click("*[type=submit]")
        page.wait_for_load_state()

        if password:
            page.locator("input[name=password]").fill(user)
            page.locator("*[type=submit]").click()
            page.wait_for_load_state()

        contexts[user] = context.storage_state()
    return context


def admin_login(browser, url, color_scheme):
    return context_login(browser, url, color_scheme, "admin")


def user_login(browser, url, color_scheme):
    return context_login(browser, url, color_scheme, "user")


def james_login(browser, url, color_scheme):
    return context_login(browser, url, color_scheme, "james", password=False)



from flask import Blueprint
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import request
from flask import url_for

from canaille.app.i18n import gettext as _
from canaille.app.session import get_login_history
from canaille.app.session import logout_user
from canaille.app.session import remove_from_login_history
from canaille.app.templating import render_template
from canaille.core.auth import AuthenticationSession
from canaille.core.auth import get_user_from_login
from canaille.core.auth import login_placeholder
from canaille.core.auth import redirect_to_next_auth_step

from ..forms import LoginForm
from . import email
from . import otp
from . import password
from . import sms

bp = Blueprint("auth", __name__)
bp.register_blueprint(password.bp)
bp.register_blueprint(email.bp)
bp.register_blueprint(sms.bp)
bp.register_blueprint(otp.bp)


@bp.context_processor
def global_processor():
    return {
        "menu": False,
    }


@bp.before_request
def load_auth():
    g.auth = AuthenticationSession.load()


@bp.after_request
def save_auth(response):
    if "auth" in g and g.auth:
        g.auth.save()
    return response


@bp.route("/login", methods=("GET", "POST"))
@bp.route("/login/<username>", methods=("GET", "POST"))
def login(username=None):
    """Handle user login with optional account selection.

    When username is provided via URL, bypasses the login form and proceeds
    directly to authentication (password, OTP, etc.). Otherwise displays
    the login form with previously used accounts if available.

    :param username: Optional pre-selected username from URL path
    """
    if g.session:
        return redirect(
            url_for("core.account.profile_edition", edited_user=g.session.user)
        )

    g.auth = AuthenticationSession.update(user_name=username)
    g.auth.reset_auth_steps()

    if username:
        user = get_user_from_login(username)
        if user and not user.has_password() and current_app.features.has_smtp:
            return redirect(url_for("core.auth.password.firstlogin", user=user))

        return redirect_to_next_auth_step()

    form = LoginForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"
    form["login"].render_kw["placeholder"] = login_placeholder()

    if not request.form or form.form_control():
        return render_template(
            "core/login.html",
            form=form,
            login_history=get_login_history(),
        )

    user = get_user_from_login(form.login.data)
    if user and not user.has_password() and current_app.features.has_smtp:
        return redirect(url_for("core.auth.password.firstlogin", user=user))

    if not form.validate():
        logout_user()
        flash(_("Login failed, please check your information"), "error")
        return render_template(
            "core/login.html", form=form, login_history=get_login_history()
        )

    g.auth = AuthenticationSession.update(
        user_name=form.login.data, remember=form.remember.data
    )
    return redirect_to_next_auth_step()


@bp.route("/logout")
def logout():
    if user := g.session and g.session.user:
        current_app.logger.security(f"Logout {user.identifier}")

        flash(
            _(
                "You have been disconnected. See you next time %(user)s",
                user=user.name,
            ),
            "success",
        )
        logout_user()
    return redirect("/")


@bp.route("/forget/<username>")
def forget(username):
    """Remove a user from login history."""
    remove_from_login_history(username)
    return redirect(url_for(".login"))



import datetime
import logging

import pytest
import time_machine

from canaille.app import mask_email
from canaille.core.auth import AuthenticationSession
from canaille.core.mails import send_one_time_password_mail
from canaille.core.models import OTP_VALIDITY
from canaille.core.models import SEND_NEW_OTP_DELAY


@pytest.fixture
def configuration(configuration):
    configuration["CANAILLE"]["AUTHENTICATION_FACTORS"] = ["email"]
    return configuration


def test_no_auth_session_not_logged_in(testclient, user):
    """Non-logged users should not be able to access the email auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/email", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with.")
    ]


def test_no_auth_session_logged_in(testclient, logged_user):
    """Logged users should not be able to access the email auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/email", status=302)
    assert res.location == "/"


def test_not_email_step(testclient, user):
    """Users reaching the email passcode form while this is not the right auth step in their flow should be redirected there."""
    with testclient.session_transaction() as session:
        session["auth"] = AuthenticationSession(
            user_name="user",
            remaining=["password", "email"],
        ).serialize()

    res = testclient.get("/auth/email", status=302)
    assert res.location == "/auth/password"


def test_signin_with_email_otp(smtpd, testclient, backend, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    assert (
        "canaille",
        logging.SECURITY,
        "Sent one-time passcode for user to john@doe.test",
    ) in caplog.record_tuples

    backend.reload(user)

    assert len(smtpd.messages) == 1
    email_content = str(smtpd.messages[0].get_payload()[0]).replace("=\n", "")
    assert user.one_time_password in email_content

    res.form["otp"] = user.one_time_password
    res = res.form.submit(status=302, name="action", value="confirm")

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful email code authentication for user",
    ) in caplog.record_tuples

    res = res.follow()
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for user",
    ) in caplog.record_tuples


def test_signin_wrong_email_otp(testclient, user, caplog):
    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    assert (
        "canaille",
        logging.SECURITY,
        "Sent one-time passcode for user to john@doe.test",
    ) in caplog.record_tuples

    res.form["otp"] = "123456"
    res = res.form.submit(status=200, name="action", value="confirm")

    assert (
        "error",
        "The passcode you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed email code authentication for user",
    ) in caplog.record_tuples


def test_expired_email_otp(testclient, user, caplog):
    """Good expired passcode should be refused."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        with testclient.session_transaction() as session:
            assert not session.get("sessions")

        res = testclient.get("/login", status=200)

        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

        otp = user.generate_sms_or_mail_otp()
        send_one_time_password_mail(user.preferred_email, otp)
        res.form["otp"] = user.one_time_password

        traveller.shift(datetime.timedelta(seconds=OTP_VALIDITY))
        res = res.form.submit(status=200, name="action", value="confirm")

        assert (
            "error",
            "The passcode you entered is invalid. Please try again",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Failed email code authentication for user",
        ) in caplog.record_tuples


def test_send_new_mail_otp_too_early(smtpd, testclient, backend, user, caplog):
    """Check that clicking on the resend button too early should not send a new mail."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)
        assert len(smtpd.messages) == 1

        res = res.form.submit(status=302, name="action", value="resend")

        assert (
            "danger",
            f"Too many attempts. Please try again in {SEND_NEW_OTP_DELAY} seconds.",
        ) in res.flashes
        assert len(smtpd.messages) == 1
        res = res.follow(status=200)


def test_send_new_mail_otp_on_time(smtpd, testclient, backend, user, caplog):
    """Check that clicking on the resend button after an adequate delay send a new code."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)
        assert len(smtpd.messages) == 1

    with time_machine.travel("2020-01-01 01:00:11+00:00", tick=False):
        res = res.form.submit(status=302, name="action", value="resend")

        assert len(smtpd.messages) == 2
        email_content = str(smtpd.messages[1].get_payload()[0]).replace("=\n", "")

        backend.reload(user)
        assert user.one_time_password in email_content
        assert (
            "success",
            "The new verification code have been sent.",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Sent one-time passcode for user to john@doe.test",
        ) in caplog.record_tuples

        assert res.location == "/auth/email"


def test_send_new_mail_invalid_user(smtpd, testclient, backend, user, caplog):
    """Invalid users should be shown the re-send button, but it should do nothing."""
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "invalid"
        res = res.form.submit(status=302)
        res = res.follow(status=200)
        assert len(smtpd.messages) == 0

    with time_machine.travel("2020-01-01 01:00:11+00:00", tick=False):
        res = res.form.submit(status=302, name="action", value="resend")

        assert len(smtpd.messages) == 0
        assert (
            "success",
            "The new verification code have been sent.",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Sent one-time passcode for user to john@doe.test",
        ) not in caplog.record_tuples


def test_send_new_email_error(smtpd, testclient, backend, user, caplog):
    """Handle SMS sending errors."""
    testclient.app.config["CANAILLE"]["SMTP"]["HOST"] = "invalid.test"

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False):
        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

    with time_machine.travel("2020-01-01 01:01:00+00:00", tick=False):
        res = res.form.submit(status=302, name="action", value="resend")
        assert (
            "canaille",
            logging.WARNING,
            "Could not send email: [Errno -2] Name or service not known",
        ) in caplog.record_tuples


def test_no_flash_when_cannot_send_new_otp_on_first_access(
    testclient, backend, user, caplog, smtpd
):
    """Test that the if flashes: condition at line 80 can be false.

    This happens when send_email_otp(flashes=False) is called from line 38
    but can_send_new_otp() returns False because an OTP was sent too recently.
    """
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        user.generate_sms_or_mail_otp()
        backend.save(user)

        # Move time forward by 5 seconds (less than SEND_NEW_OTP_DELAY of 10 seconds)
        traveller.shift(datetime.timedelta(seconds=5))

        res = testclient.get("/login", status=200)
        res.form["login"] = "user"
        res = res.form.submit(status=302)

        # At this point:
        # - g.auth.current_step_start_dt is at T+5 (just now)
        # - user.one_time_password_emission_date is at T (5 seconds ago)
        # So g.auth.current_step_start_dt > user.one_time_password_emission_date triggers line 38
        # But can_send_new_otp() returns False (only 5 seconds passed < 10 seconds)
        res = res.follow(status=200)

        # No email sent because can_send_new_otp() returned False
        assert len(smtpd.messages) == 0

        # No flash message because flashes=False at line 38
        assert res.flashes == []

        assert "Email verification code" in res.text


def test_mask_email():
    email = "foo@bar.com"
    assert mask_email(email) == "f#####o@bar.com"

    email = "hello"
    assert mask_email(email) is None



from flask import Blueprint
from flask import jsonify
from flask import request

from ..oauth import oauth_authorization_server
from ..oauth import openid_configuration

bp = Blueprint("home", __name__, url_prefix="/.well-known")


@bp.route("/oauth-authorization-server")
def oauth_authorization_server_endpoint():
    return jsonify(
        {
            key: val
            for key, val in oauth_authorization_server().items()
            if val is not None
        }
    )


@bp.route("/openid-configuration")
def openid_configuration_endpoint():
    return jsonify(
        {key: val for key, val in openid_configuration().items() if val is not None}
    )


@bp.route("/webfinger")
def webfinger():
    return jsonify(
        {
            "links": [
                {
                    "href": openid_configuration()["issuer"],
                    "rel": "http://openid.net/specs/connect/1.0/issuer",
                }
            ],
            "subject": request.args["resource"],
        }
    )



import json

from canaille.app import models
from canaille.commands import cli


def test_restore_stdin(cli_runner, backend):
    """Test the full database dump command."""
    payload = {
        "group": [
            {
                "created": "2025-01-01T12:00:00+00:00",
                "display_name": "foo",
                "id": "cef66b33-2762-423e-b0c3-2fb6ab13abfa",
                "last_modified": "2025-01-01T12:00:00+00:00",
                "members": [
                    "e52b36b5-6a94-4395-b3a8-0f72d9140bfa",
                ],
            },
        ],
        "user": [
            {
                "created": "2025-01-02T12:00:00+00:00",
                "display_name": "Johnny",
                "emails": [
                    "john@doe.test",
                ],
                "family_name": "Doe",
                "formatted_address": "1235, somewhere",
                "formatted_name": "John (johnny) Doe",
                "given_name": "John",
                "groups": ["cef66b33-2762-423e-b0c3-2fb6ab13abfa"],
                "id": "e52b36b5-6a94-4395-b3a8-0f72d9140bfa",
                "last_modified": "2025-01-02T12:00:00+00:00",
                "password": "$pbkdf2-sha512$25000$Nsa4VwoBgHBOKaUUImSMsQ$/Ynkpp1RQILSMBUfKxxMZpG/2oYcFNJxFGMZ0xoHSIbIGkkqeo.VgY71r3Gl8tepUJRmNnaW7if0C5pRHue/Zw",
                "phone_numbers": [
                    "555-000-000",
                ],
                "preferred_language": "en",
                "profile_url": "https://john.test",
                "user_name": "user",
            },
        ],
    }

    res = cli_runner.invoke(cli, ["restore"], input=json.dumps(payload))
    assert res.exit_code == 0, res.stdout

    user = backend.get(models.User)
    assert user.user_name == "user"
    assert user.family_name == "Doe"

    group = backend.get(models.Group)
    assert group.display_name == "foo"

    assert user.groups == [group]
    assert group.members == [user]

    backend.delete(group)
    backend.delete(user)


def test_restore_stdin_no_input(cli_runner, backend):
    """Test the restore command without input."""
    res = cli_runner.invoke(cli, ["restore"], catch_exceptions=False)
    assert res.exit_code == 1, res.stdout


def test_restore_stdin_empty_input(cli_runner, backend):
    """Test the restore command with an empty input."""
    res = cli_runner.invoke(cli, ["restore"], input="")
    assert res.exit_code == 1, res.stdout


def test_restore_stdin_invalid_input(cli_runner, backend):
    """Test the restore command with an invalid input."""
    res = cli_runner.invoke(cli, ["restore"], input="invalid")
    assert res.exit_code == 1, res.stdout



import json
import logging
from unittest import mock

import pytest

from canaille.commands import cli


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_reset_otp_by_id(testclient, cli_runner, backend, caplog, user, otp_method):
    """Reset one-time passcode authentication for a user by its id."""
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    old_token = user.secret_token
    assert old_token is not None

    res = cli_runner.invoke(
        cli,
        [
            "reset-otp",
            user.id,
        ],
    )
    assert res.exit_code == 0, res.stderr
    assert json.loads(res.stdout) == {
        "created": mock.ANY,
        "display_name": "Johnny",
        "emails": [
            "john@doe.test",
        ],
        "family_name": "Doe",
        "formatted_address": "1234, some street, 6789 some city, some state",
        "formatted_name": "John (johnny) Doe",
        "given_name": "John",
        "id": user.id,
        "last_modified": mock.ANY,
        "password": mock.ANY,
        "phone_numbers": [
            "555-000-000",
        ],
        "preferred_language": "en",
        "profile_url": "https://john.test",
        "user_name": "user",
        "hotp_counter": 1,
        "secret_token": mock.ANY,
        "locality": "some city",
        "postal_code": "6789",
        "region": "some state",
        "street": "1234, some street",
        "last_otp_login": "2020-01-01T00:00:00+00:00",
    }
    assert (
        "canaille",
        logging.SECURITY,
        "Reset one-time passcode authentication from CLI for user",
    ) in caplog.record_tuples
    backend.reload(user)
    assert user.secret_token is not None
    assert user.secret_token != old_token
    if otp_method == "HOTP":
        assert user.hotp_counter == 1


def test_reset_otp_unknown_id(cli_runner):
    """Error case for trying to reset one-time passcode authentication for an invalid user."""
    res = cli_runner.invoke(
        cli,
        [
            "reset-otp",
            "invalid",
        ],
    )
    assert res.exit_code == 1, res.stdout
    assert res.stderr == "Error: No user with id 'invalid'\n"
